# basic-level BHAV decoding
#
# - functions
#
# get_goto(package, instance, byte-type): get the name of a specified transition
# get_op(package, opcode, buffer, index, format, node_ver): get listing of a specific op
# get_xrefs(package): get list of external references to current BHAV
# get_tree(package, dump): get listing of current BHAV group/instance (None = not found)

import dz_pack
import dz_prim
import dz_util
from dz_defs import *
from dz_glob import *

def get_goto(p, i, b):
    """Return the name of a specified transition."""
    if b:
        if i == 0xFD:
            return -3, dz_util.get_html(p,TYPE_GOTO_CODE,"error")  # i.e. raise undefined transition error
        elif i == 0xFE:
            return -2, dz_util.get_html(p,TYPE_GOTO_CODE,"true")   # i.e. pop this frame, goto callee's 'true' node
        elif i == 0xFF:
            return -1, dz_util.get_html(p,TYPE_GOTO_CODE,"false")  # i.e. pop this frame, goto callee's 'false' node
        else:
            return i, dz_util.get_html(p,TYPE_GOTO_CODE,"%X" % i)
    else:
        if i == 0xFFFC:
            return -3, dz_util.get_html(p,TYPE_GOTO_CODE,"error")
        elif i == 0xFFFD:
            return -2, dz_util.get_html(p,TYPE_GOTO_CODE,"true")
        elif i == 0xFFFE:
            return -1, dz_util.get_html(p,TYPE_GOTO_CODE,"false")
        else:
            return i, dz_util.get_html(p,TYPE_GOTO_CODE,"%X" % i)

def get_op(p, opcode, b, x, f, node_version):
    """Return a string for a specified operation."""
    if opcode < len(gString8B):
        return dz_prim.get_prim(p, opcode, b, x, f, node_version)
    else:
        # not primitive, get tree name and params, etc.
        d = dz_prim.get_title(p, opcode)
        s = ""

        # check objects if not found in this package
        if not d and not p.is_objects and p.objects:
            p2 = p.objects
            p2.group = p.group
            p2.instance = p.instance
            p2.semi = p.semi
            d = dz_prim.get_title(p2, opcode)

        # parse params
        if d:
            s += dz_util.get_html(p,TYPE_LINK_FUNCTION,d['name'],(d['group'],opcode))
            k = dz_util.s2i(b[x:x+4])
            i = d['params']
            if i > 0 and k != 0xFFFFFFFF and (x + 15) < len(b):
                w = 0
                c = ord(b[x+12])
                s += "("
                if (c & 1) != 0 and c != 0xFF:
                    for j in range(i):
                        if j != 0:
                            s += ", "
                        v = dz_util.s2i(b[x+w+1:x+w+3])
                        s += dz_prim.get_data(p, ord(b[x+w]), v)    # (owner, value)
                        w += 3
                elif (c & 2) != 0 and c != 0xFF and node_version != 0:
                    for j in range(i):
                        if j != 0:
                            s += ", "
                        s += "original param %d" % j
                else:
                    for j in range(i):
                        if j != 0:
                            s += ", "
                        v = dz_util.s2i(b[x+w:x+w+2])
                        s += dz_util.ext16(v)
                        w += 2
                s += ")"
        else:
            s += "[BHAV 0x%X]" % opcode
        p.markup.append( (1, opcode, p.m_offset, p.m_offset + len(s)) )
        return s

def get_xrefs(p):
    """Return a list of instances in this group that refer to the current BHAV."""
    this_i = p.instance
    g = p.group
    out = []

    # check the cache (A+ optimization for dump mode)
    if 'group' in p.xref_cache:
        if p.xref_cache['group'] == g:
            for i,j in p.xref_cache.iteritems():
                if i != 'group' and this_i in j:
                    out.append(i)
            return out

    # this group is not in the cache, rebuild it
    p.xref_cache = {'group': g}
    r = p.find({'type': TYPE_BHAV, 'group': g})
    while r:
        b = p.get(r['offset'], r['size'])
        if b:
            i = r['instance']
            f = dz_util.s2i(b[0x40:0x42])   # format
            num = dz_util.s2i(b[0x42:0x44]) # number of opcodes
            x = 0x4C                        # x = next index
            if f > 0x8008:
                x = 0x4D

            for n2 in range(num):
                if x >= len(b):     # range check
                    break           # terminate n2 loop

                op = dz_util.s2i(b[x:x+2])
                if op >= 0x100:
                    if i not in p.xref_cache:
                        p.xref_cache[i] = [op]
                    else:
                        p.xref_cache[i].append(op)  # todo: fix redundancy here
                if op == this_i and op not in out and i != this_i:
                    out.append(i)

                if f < 0x8007:
                    if f < 0x8003:
                        x += 12             # 4 ins bytes + 8 operand bytes
                    elif f < 0x8005:
                        x += 20             # 4 ins bytes + 16 operand bytes
                    else:
                        x += 21             # 5 ins bytes + 16 operand bytes
                else:
                    x += 23                 # 7 ins bytes + 16 operand bytes

        r = p.find({'type': TYPE_BHAV, 'group': g, 'index': r['index']})

    return out

def get_tree(p, dump):
    """Return the disassembly listing for the current group/instance BHAV."""
    g = p.group
    i = p.instance
    p.markup = []                   # reset current markup
    p.m_offset = 0

    r = p.find({'type': TYPE_BHAV, 'group': g, 'instance': i})
    if not r: return "[BHAV 0x%X]" % i  # not found

    b = p.get(r['offset'], r['size'])
    if not b: return b              # inexplicably missing (!)

    out = []                        # listing
    out2 = []                       # data dump (if used)
    out3 = []                       # markup result
    trans_t = []                    # transitions
    trans_f = []

    j = 0x40                        # get title (cstring, maxchars = 0x40 at index = 0)
    if '\x00' in b[:0x40]: j = b.index('\x00')
    out.append(dz_util.get_html(p,TYPE_COMMENT,"# Title = %s\n#\n" % b[:j]))
    out2.append("")

    # get header
    f = dz_util.s2i(b[0x40:0x42])   # format
    num = dz_util.s2i(b[0x42:0x44]) # number of opcodes
    c0 = ord(b[0x44])               # tree type
    c1 = ord(b[0x45])               # number of args (params)
    c2 = ord(b[0x46])               # number of locals
    c3 = ord(b[0x47])               # header flag
    d1 = dz_util.s2i(b[0x48:0x4C])  # tree version
    out.append( dz_util.get_html(p,TYPE_COMMENT,"# Format = 0x%X, Params = %d, Locals = %d\n" % (f, c1, c2)) )
    out2.append("")

    d2 = d1
    if d1 > 0x7FFFFFFF:
        d2 = d1 - 0x100000000       # sign-extend 32-bit value
    s = dz_util.get_html(p,TYPE_COMMENT,"# Tree type = 0x%X, Header flag = 0x%X, Tree version = 0x%X (%d)" \
        % (c0, c3, d1, d2))
    x = 0x4C                        # x = next index
    if f > 0x8008:
        s += dz_util.get_html(p,TYPE_COMMENT,", Cache flags = %d" % ord(b[0x4C]))
        x = 0x4D
    out.append(s + "\n\n")
    out2.append("")

    # get instruction listing
    for n2 in range(num):
        p.markup = []       # reset current markup
        p.m_offset = 0

        if x >= len(b):     # range check
            out.append("*** Error: too many instructions (%X) ***\n" % num)
            break           # break out of for loop

        # line num, get opcode
        s = dz_util.get_html(p,TYPE_CODE_LINE,"%6X: " % n2)
        op = dz_util.s2i(b[x:x+2])
        p.m_offset = len(s)

        # dump data, if specified
        if dump:
            k = 8   # 16 bytes of operands
            if f == 0x8004:
                x2 = 4
            elif f == 0x8005 or f == 0x8006:
                x2 = 5
            elif f > 0x8006:
                x2 = 7
            else:
                x2 = 4
                k = 4   # 8 bytes of operands
            s2 = "        [ %04X: " % op
            s2 += "(%02X) " % ord(b[x+x2-1])    # node version
            for j in range(k):
                s2 += "%02X%02X" % ( ord(b[x+x2]), ord(b[x+x2+1]) )
                x2 += 2
            s2 += " ]\n"
            out2.append(s2)

        # continue listing
        if f < 0x8007:              # byte-size transitions
            c1 = ord(b[x+2])        # true
            c2 = ord(b[x+3])        # false
            if f < 0x8005:
                s += get_op(p, op, b, x+4, f, 0)
            else:
                s += get_op(p, op, b, x+5, f, ord(b[x+4]))
            s += "; true: "
            gt, gs = get_goto(p, c1, True)
            s += gs
            trans_t.append(gt)
            s += ", false: "
            gt, gs = get_goto(p, c2, True)
            s += gs + "\n"
            trans_f.append(gt)

            if f < 0x8003:
                x += 12             # 4 ins bytes + 8 operand bytes
            elif f < 0x8005:
                x += 20             # 4 ins bytes + 16 operand bytes
            else:
                x += 21             # 5 ins bytes + 16 operand bytes
        else:
            w1 = dz_util.s2i(b[x+2:x+4])    # true
            w2 = dz_util.s2i(b[x+4:x+6])    # false
            s += get_op(p, op, b, x+7, f, ord(b[x+6]))
            s += "; true: "
            gt, gs = get_goto(p, w1, False)
            s += gs
            trans_t.append(gt)
            s += ", false: "
            gt, gs = get_goto(p, w2, False)
            s += gs + "\n"
            trans_f.append(gt)
            x += 23                 # 7 ins bytes + 16 operand bytes

        out.append(s)
        out3.append( (0, len(out) - 1, 0, 0) )  # mt = 0: markup header (line index)
        for j in range(len(p.markup)):
            out3.append(p.markup[j])            # + current markup list

    # end of for n2 loop, return listing(s)

    x = get_xrefs(p)

    # output for xrefs
    s = "\n"
    for i in range(len(x)):
        s += "0x%X" % x[i]
        if i+1 < len(x):
            s += ", "
    if len(x):
        out.append("\n# This is called by:\n")
        out.append(s+"\n")

    out.append("\n# Paths:\n")
    r_path = []
    for i in range(len(trans_t)):
        s = []
        for j in range(len(trans_t)):
            if trans_t[j] == i or trans_f[j] == i:
                s.append(j)
        r_path.append(s)        # reverse references

    # get linear/non-linear/unreferenced paths
    a_path = []
    b_path = []
    u_path = []
    for i in range(len(r_path)):
        if len(r_path[i]) == 1:
            a_path.append(r_path[i][0])       # linear from, to
            a_path.append(i)
        elif len(r_path[i]) == 0:
            u_path.append(i)                  # outside path
        else:
            b_path.append(r_path[i])          # non-linear from, to
            b_path.append(i)

    # sort linear paths
    a2a_path = []
    a2b_path = []
    i = 0
    while i < len(a_path):
        x = False
        for j in range(len(a2a_path)):
            if a_path[i] < a2a_path[j]:        # sort by from
                a2a_path.insert(j, a_path[i])
                a2b_path.insert(j, a_path[i+1])
                x = True
                break   # terminate j loop
        if not x:
            a2a_path.append(a_path[i])
            a2b_path.append(a_path[i+1])
        i += 2

    # get compressed linear paths
    ca_path = []
    cb_path = []
    for i in range(len(a2a_path)):
        if a2b_path[i] in a2a_path:            # to (i) = from (x)?
            x = a2a_path.index(a2b_path[i])
            if a2a_path[i] not in ca_path:
                ca_path.append(a2a_path[i])    # from (i)
                cb_path.append(a2b_path[i])    # to (i)
            if a2a_path[x] not in ca_path:
                ca_path.append(a2a_path[x])    # from (x)
                cb_path.append(a2b_path[x])    # to (x)

    # sort compressed linear paths
    c_path = []
    for i in range(len(ca_path)):
        if ca_path[i] not in c_path:           # needs from?
            if cb_path[i] in c_path:
                x = c_path.index(cb_path[i])
                c_path.insert(x, ca_path[i])   # insert before to
            else:
                c_path.append(ca_path[i])

        if cb_path[i] not in c_path:           # needs to?
            if ca_path[i] in c_path:
                x = c_path.index(ca_path[i])
                c_path.insert(x+1, cb_path[i]) # insert after from
            else:
                c_path.append(cb_path[i])

    # insert compressed linear path breaks
    i = 0
    while i+1 < len(c_path):
        if c_path[i] in ca_path:
            x = ca_path.index(c_path[i])
            if c_path[i+1] != cb_path[x]:
                c_path.insert(i+1, -1)
                i += 1
        else:
            c_path.insert(i+1, -1)
            i += 1
        i += 1

    # output for compressed linear paths
    s = "\n"
    for i in range(len(c_path)):
        if c_path[i] >= 0:
            s += "%X" % c_path[i]
            if i+1 < len(c_path) and c_path[i+1] >= 0:
                s += " > "
            if c_path[i] in u_path:
                u_path.remove(c_path[i])    # remove from unreferenced
        else:
            if i+1 < len(c_path):
                s += "\n"
    if len(c_path): out.append(s+"\n")

    # output for linear paths
    i = 0
    x = False
    s = "\n"
    for i in range(len(a2a_path)):
        if a2a_path[i] not in ca_path or a2b_path[i] not in cb_path:
            x = True
            s += "%X > %X; " % (a2a_path[i], a2b_path[i])
            if a2a_path[i] in u_path:
                u_path.remove(a2a_path[i])      # if line is a from, remove from unreferenced
    if x: out.append(s[:-2]+"\n")

    # output for non-linear paths
    i = 0
    s = "\n"
    while i < len(b_path):
        x = b_path[i]
        for j in range(len(x)):
            s += "%X" % x[j]
            if x[j] in u_path:
                u_path.remove(x[j])      # if line is a from, remove from unreferenced
            if j+1 < len(x):
                s += ", "
        s += " > %X" % b_path[i+1]
        i += 2
        if i < len(b_path): s += "\n"
    if len(b_path): out.append(s+"\n")

    # output for unreferenced
    s = "\nno refs: "
    for i in range(len(u_path)):
        s += "%X " % u_path[i]
    if len(u_path): out.append(s+"\n")

    if dump:
        return (out, out2, out3)
    else:
        return (out, out3)
