# Lua decoding
#
# - functions:
#
# get_int(buffer, index): return integer value at index
# get_ints(buffer, index): return list of integers at index
# get_string(buffer, index): return string value at index
# get_strings(buffer, index): return list of strings at index
# get_locals(buffer, index): return list of locals at index
# get_consts(buffer, index): return list of constants at index
# get_rk(operand, constants): return a string for register or constant
# get_instr(buffer, index, constants, line, level): return decoded instruction at index
# get_fchunk(buffer, index, level, f_index): decode the current Lua function chunk
# get_chunk(package, type): decode the current Lua chunk

import struct

import dz_pack
import dz_util

def get_int(b, x):
    i = dz_util.s2i(b[x:x+4])
    return i, x+4

def get_ints(b, x):
    n, x = get_int(b, x)
    s = []
    for i in range(n):
        c, x = get_int(b, x)
        s.append(c)
    return s, x

def get_string(b, x):
    n, x = get_int(b, x)                # note: n should never be 0
    s = b[x:x+n-1]                      # -1 to exclude null-terminator
    return s, x+n

def get_strings(b, x):
    n, x = get_int(b, x)
    s = []
    for i in range(n):
        s2, x = get_string(b, x)
        s.append(s2)
    return s, x

def get_locals(b, x):
    n, x = get_int(b, x)
    s = []
    for i in range(n):
        s2, x = get_string(b, x)        # variable name
        c1, x = get_int(b, x)           # start pc
        c2, x = get_int(b, x)           # end pc
        s.append([s2, c1, c2])
    return s, x

def get_consts(b, x):
    n, x = get_int(b, x)
    s = []
    for i in range(n):
        t = ord(b[x])                   # type of constant
        x += 1
        if t == 3:                      # number
            xd = struct.unpack('<d', b[x:x+8])  # simple double to float conversion
            x += 8
            c = xd[0]
            if c == int(c):
                c = int(c)    # kill any .0s
                if c > 16:
                    c = "%d (0x%X)" % (c, c)
                elif c < -16:
                    c = "%d (0x%X)" % (c, 0x100000000+c)  # note: 32-bit values are assumed here
        elif t == 4:                    # string
            c, x = get_string(b, x)
            c = "'" + c + "'"
        else:
            c = 0
        s.append(c)
    return s, x

def get_rk(op, constants):
    if op >= 250:
        s = str(constants[op-250])
    else:
        s = "r%d" % op
    return s

def get_instr(b, x, constants, line, level):
    ins, x = get_int(b, x)
    opcode = ins & 0x3F                 # aaaaaaaa bbbbbbbb bccccccc ccoooooo
    op_a = ins >> 24
    op_b = (ins >> 15) & 0x1FF
    op_c = (ins >> 6) & 0x1FF
    op_bx = (ins >> 6) & 0x3FFFF
    op_sbx = op_bx - 0x1FFFF            # bias signed

    s = "%6d: " % line

    if opcode == 0:                     # OP_MOVE
        s += "r%d := r%d" % (op_a, op_b)
    elif opcode == 1:                   # OP_LOADK
        s += "r%d := %s" % (op_a, str(constants[op_bx]))
    elif opcode == 2:                   # OP_LOADBOOL
        s += "r%d := %d" % (op_a, op_b)
        if op_c != 0:
            s += "; goto %d" % (line + 2)
    elif opcode == 3:                   # OP_LOADNIL
        if op_a == op_b:
            s += "r%d := nil" % op_a
        else:
            s += "r%d to r%d := nil" % (op_a, op_b)
    elif opcode == 4:                   # OP_GETUPVAL
        s += "r%d := up_value[%d]" % (op_a, op_b)
    elif opcode == 5:                   # OP_GETGLOBAL
        s += "r%d := global[%s]" % (op_a, str(constants[op_bx]))
    elif opcode == 6:                   # OP_GETTABLE
        s += "r%d := r%d[%s]" % (op_a, op_b, get_rk(op_c, constants))
    elif opcode == 7:                   # OP_SETGLOBAL
        s += "global[%s] := r%d" % (str(constants[op_bx]), op_a)
    elif opcode == 8:                   # OP_SETUPVAL
        s += "up_value[%d] := r%d" % (op_b, op_a)
    elif opcode == 9:                   # OP_SETTABLE
        s += "r%d[%s] := %s" % (op_a, get_rk(op_b, constants), get_rk(op_c, constants))
    elif opcode == 0xA:                 # OP_NEWTABLE
        s += "r%d := {} (size = %d, %d)" % (op_a, op_b, op_c)
    elif opcode == 0xB:                 # OP_SELF
        s += "r%d := r%d; r%d := r%d[%s]" % (op_a+1, op_b, op_a, op_b, get_rk(op_c, constants))

    elif opcode == 0xC:                 # OP_ADD
        s += "r%d := %s + %s" % (op_a, get_rk(op_b, constants), get_rk(op_c, constants))
    elif opcode == 0xD:                 # OP_SUB
        s += "r%d := %s - %s" % (op_a, get_rk(op_b, constants), get_rk(op_c, constants))
    elif opcode == 0xE:                 # OP_MUL
        s += "r%d := %s * %s" % (op_a, get_rk(op_b, constants), get_rk(op_c, constants))
    elif opcode == 0xF:                 # OP_DIV
        s += "r%d := %s / %s" % (op_a, get_rk(op_b, constants), get_rk(op_c, constants))
    elif opcode == 0x10:                # OP_POW
        s += "r%d := %s ^ %s" % (op_a, get_rk(op_b, constants), get_rk(op_c, constants))
    elif opcode == 0x11:                # OP_UNM
        s += "r%d := -r%d" % (op_a, op_b)
    elif opcode == 0x12:                # OP_NOT
        s += "r%d := not r%d" % (op_a, op_b)
    elif opcode == 0x13:                # OP_CONCAT
        s += "r%d := concat r%d to r%d" % (op_a, op_b, op_c)

    elif opcode == 0x14:                # OP_JMP
        s += "goto %d" % (line + op_sbx + 1)

    elif opcode == 0x15:                # OP_EQ
        if not op_a:
            s += "if %s == %s then goto %d" % (get_rk(op_b, constants), get_rk(op_c, constants), line+2)
        else:
            s += "if %s != %s then goto %d" % (get_rk(op_b, constants), get_rk(op_c, constants), line+2)
    elif opcode == 0x16:                # OP_LT
        if not op_a:
            s += "if %s < %s then goto %d" % (get_rk(op_b, constants), get_rk(op_c, constants), line+2)
        else:
            s += "if %s >= %s then goto %d" % (get_rk(op_b, constants), get_rk(op_c, constants), line+2)
    elif opcode == 0x17:                # OP_LE
        if not op_a:
            s += "if %s <= %s then goto %d" % (get_rk(op_b, constants), get_rk(op_c, constants), line+2)
        else:
            s += "if %s > %s then goto %d" % (get_rk(op_b, constants), get_rk(op_c, constants), line+2)
    elif opcode == 0x18:                # OP_TEST
        s += "if "
        if op_c: s += "not "
        s += "r%d then goto %d" % (op_b, line+2)
        if op_b != op_a:
            s += " else r%d := r%d" % (op_a, op_b)

    elif opcode == 0x19:                # OP_CALL
        if op_c == 0:
            s += "r%d ... := " % op_a
        elif op_c == 2:
            s += "r%d := " % op_a
        elif op_c != 1:
            s += "r%d to r%d := " % (op_a, op_a+op_c-2)
        s += "call r%d" % op_a
        if op_b == 0:
            s += "(r%d ... )" % (op_a+1)
        elif op_b == 1:
            s += "()"
        elif op_b == 2:
            s += "(r%d)" % (op_a+1)
        elif op_b == 3:
            s += "(r%d, r%d)" % (op_a+1, op_a+2)
        else:
            s += "(r%d to r%d)" % (op_a+1, op_a+op_b-1)
    elif opcode == 0x1A:                # OP_TAILCALL
        s += "return call r%d" % op_a
        if op_b == 0:
            s += "(r%d ... )" % (op_a+1)
        elif op_b == 1:
            s += "()"
        elif op_b == 2:
            s += "(r%d)" % (op_a+1)
        elif op_b == 3:
            s += "(r%d, r%d)" % (op_a+1, op_a+2)
        else:
            s += "(r%d to r%d)" % (op_a+1, op_a+op_b-1)
    elif opcode == 0x1B:                # OP_RETURN
        s += "return"
        if op_b == 0:
            s += " r%d ..." % op_a
        elif op_b == 2:
            s += " r%d" % op_a
        elif op_b == 3:
            s += " r%d, r%d" % (op_a, op_a+1)
        elif op_b != 1:
            s += " r%d to r%d" % (op_a, op_a+op_b-2)

    elif opcode == 0x1C:                # OP_FORLOOP
        s += "r%d := r%d + r%d; " % (op_a, op_a, op_a+2)
        s += "if r%d <= r%d then goto %d" % (op_a, op_a+1, line + op_sbx + 1)
    elif opcode == 0x1D:                # OP_TFORLOOP
        if op_c == 0:
            s += "r%d := " % (op_a+2)
        elif op_c == 1:
            s += "r%d, r%d := " % (op_a+2, op_a+3)
        else:
            s += "r%d to r%d := " % (op_a+2, op_a+2+op_c)
        s += "call r%d(state=r%d, index=r%d); " % (op_a, op_a+1, op_a+2)
        s += "if not r%d then goto %d" % (op_a+2, line+2)
    elif opcode == 0x1E:                # OP_TFORPREP (note: Lua v5 is assumed here)
        s += "goto %d" % (line + op_sbx + 1)

    elif opcode == 0x1F:                # OP_SETLIST
        mbx = op_bx & 0x1F        # Bx%FPF
        s += "r%d[%d to %d] := r%d to r%d" % (op_a, op_bx-mbx+1, op_bx+1, op_a+1, op_a+mbx+1)
    elif opcode == 0x20:                # OP_SETLISTO
        mbx = op_bx & 0x1F        # Bx%FPF
        s += "r%d[%d ...] := r%d ..." % (op_a, op_bx-mbx+1, op_a+1)

    elif opcode == 0x21:                # OP_CLOSE
        s += "close r%d ..." % op_a
    elif opcode == 0x22:                # OP_CLOSURE (note: doesn't account for upvalues)
        s += "r%d := Function %d (Level=%d)" % (op_a, op_bx, level+1)

    s += "\n"
    return s, x

def get_fchunk(b, x, level, fx):
    out = []                            # listing for this function

    source_name, x = get_string(b, x)   # usually "=(none)" or "", note: ignored
    line_def, x = get_int(b, x)         # note: ignored

    num_ups = ord(b[x])
    num_params = ord(b[x+1])
    var_arg = ord(b[x+2])
    stack_size = ord(b[x+3])
    x += 4

    source_lines, x = get_ints(b, x)    # usually [], note: ignored
    locals, x = get_locals(b, x)        # usually [], note: ignored
    up_values, x = get_strings(b, x)    # usually [], note: ignored
    constants, x = get_consts(b, x)

    # recurse through any function chunks defined in this function
    f, x = get_int(b, x)                # number of function protos for this function
    for i in range(f):
        s, x = get_fchunk(b, x, level+1, i)
        for j in range(len(s)):
            out.append(s[j])            # append returned listing to this listing

    # function header
    out.append("\n###\n# Function %d, Level %d" % (fx, level))
    if num_ups:
        out.append(", Ups = %d" % num_ups)
    if var_arg:
        out.append(", VarArg = %d" % var_arg)
    out.append(", StackSize = %d" % stack_size)
    if num_params == 1:
        out.append(", Param = r0")
    elif num_params > 1:
        out.append(", Params = r0 to r%d" % (num_params-1))
    out.append("\n\n")

    # instructions
    ins, x = get_int(b, x)              # number of instructions
    for i in range(ins):
        s, x = get_instr(b, x, constants, i+1, level)
        out.append(s)

    return out, x

def get_chunk(p, t):
    r = p.find({'type': t, 'group': p.group, 'instance': p.instance, 'subtype': p.subtype})
    if not r: return []                 # not found

    b = p.get(r['offset'], r['size'])
    if not b: return []                 # inexplicably missing (!)

    out = []                            # listing

    # header (note: this assumes a lot of these header fields don't change)
    x = dz_util.s2i(b[4:8])             # title length
    out.append( "# Title = %s\n" % b[8:8+x])
    x += 8                              # + pre-title size
    x += 13                             # + header5['size of C field']
    x += ord(b[x]) + 1                  # size of Lua number (usually 8) + size of byte

    # get function chunk(s)
    s, x = get_fchunk(b, x, 1, 0)       # note: assumes there's one level 1 function
    for i in range(len(s)):
        out.append(s[i])                # append returned listing to this listing
    return out
