# mutable package handling class
# -------------------------------------
# - constructor:
#
# MutablePackage(filename, prefs, new): creates (if new = True) and/or loads a package
#
# - methods:
#
# append(dict): create a new instance
# remove(dict): remove an instance
# set(dict): modify an instance
# can_undo(): get undo status
# can_redo(): get redo status
# undo(): undo a recent change
# redo(): redo a recent change
# do_change(dict): perform indicated change to package
# is_modified(dict): get status of an instance
# get(offset, size): get data, as modified
# can_write(): get write status
# write(): save changes to the package
# build_start(): start building a package (note: truncates file)
# build_append(dict): append a new instance to a package build
# build_end(): finish building a package

import logging
import struct

import dz_pack
import dz_util
from dz_defs import *

class MutablePackage(dz_pack.Package):
    def __init__(self, file_name, prefs, is_new):
        """Create or load the TOC of a package file."""
        self.change_stack = []
        self.redo_stack = []
        self.hdr_size = 0x60

        if is_new:
            b = "DBPF\x01\x00\x00\x00\x01\x00\x00\x00"  # (00 to 0B) ID + version
            b += "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"     # (0C to 17) reserved
            b += "\x00\x00\x00\x00"     # (18 to 1B) creation date
            b += "\x00\x00\x00\x00"     # (1C to 1F) modification date
            b += "\x07\x00\x00\x00"     # (20 to 23) TOC major ver
            b += "\x00\x00\x00\x00"     # (24 to 27) number of entries in TOC
            b += "\x00\x00\x00\x00"     # (28 to 2B) offset of TOC start
            b += "\x00\x00\x00\x00"     # (2C to 2F) size of TOC
            b += "\x00\x00\x00\x00"     # (30 to 33) number of hole entries
            b += "\x00\x00\x00\x00"     # (34 to 37) offset of hole start
            b += "\x00\x00\x00\x00"     # (38 to 3B) size of hole
            b += "\x02\x00\x00\x00"     # (3C to 3F) TOC minor ver (2 = include subtype)
            for i in range(32):         # (40 to 5F) reserved
                b += "\x00"

            try:
                f = open(file_name, 'wb')
                f.write(b)
            except:
                f = None
                logging.error("mpack.init: create file %s failed" % file_name)
            if f: f.close()

        dz_pack.Package.__init__(self, file_name, prefs, False)  # init Package

    def append(self, d):
        """Create a new instance."""
        d['function'] = 1               # 1 = append
        self.change_stack.append(d.copy())

    def remove(self, d):
        """Delete a specified instance."""
        d['function'] = -1              # -1 = remove
        self.change_stack.append(d.copy())

    def set(self, d):
        """Modify a specified instance."""
        d['function'] = 0               # 0 = set
        self.change_stack.append(d.copy())

    def can_undo(self):
        return len(self.change_stack) != 0

    def can_redo(self):
        return len(self.redo_stack) != 0

    def undo(self):
        """Undo a recent change."""
        self.redo_stack.append(self.change_stack.pop())

    def redo(self):
        """Redo a recent change."""
        self.change_stack.append(self.redo_stack.pop())

    def do_change(self, d, writing=True):
        """Perform indicated change to package."""
        mFunction = d['function']
        mGroup, mType, mInstance, mSubType = 0, 0, 0, 0
        mData = ""
        mRange = [0, 0]     # range to replace (i.e. [x,x] = insert, [0,0] = prepend)

        if 'group' in d:
            mGroup = d['group']
        if 'type' in d:
            mType = d['type']
        if 'instance' in d:
            mInstance = d['instance']
        if 'subtype' in d:
            mSubType = d['subtype']
        if 'data' in d:
            mData = d['data']
        if 'range' in d:
            mRange = d['range']

        d2 = dict(type=mType, group=mGroup, instance=mInstance, subtype=mSubType)

        if mFunction == 1 and writing:              # append
            if d2 not in self.mod_entries:
                self.mod_entries.append(d2)         # add entry to mod
                self.mod_data.append("")            # with empty buffer

            x3 = [mType, mInstance, mSubType, 0, 0] # offset, size = 0, 0
            if mGroup not in self.groups:
                self.groups.append(mGroup)          # new group
                self.g_data.append([x3])            # with one entry
            else:
                x4 = self.groups.index(mGroup)      # x4 = group index
                self.g_data[x4].append(x3)          # add this entry

            self.toc_num += 1
            self.toc_valid = True                   # this entry is find()able

        elif mFunction == -1 and writing:           # remove
            if d2 in self.mod_entries:
                x = self.mod_entries.index(d2)
                self.mod_entries.pop(x)             # remove entry from mod
                self.mod_data.pop(x)

            if mGroup in self.groups:
                x = self.groups.index(mGroup)
                r = dz_pack.Package.find(self, d2)
                if r:
                    j = r['index'][1]
                    self.g_data[x].pop(j-1)         # remove entry
                    if len(self.g_data[x]) == 0:    # entries = 0?
                        self.groups.pop(x)          # remove group
                        self.g_data.pop(x)

                    self.toc_num -= 1
                    if self.toc_num == 0:
                        self.toc_valid = False      # no entries = toc invalid

        elif mFunction == 0:                        # set
            # get buffer
            if d2 not in self.mod_entries:          # first set for existing entry
                b = ""
                x = len(self.mod_entries)
                self.mod_entries.append(d2)
                r = dz_pack.Package.find(self, d2)
                if r:
                    b = dz_pack.Package.get(self, r['offset'], r['size'])
                    if not b: b = ""
                self.mod_data.append(b)

            else:
                x = self.mod_entries.index(d2)
                b = self.mod_data[x]

            # apply set
            self.mod_data[x] = b[:mRange[0]] + mData + b[mRange[1]:]

    def is_modified(self, d):
        """Return whether this instance has been modified."""
        # get current modifications
        self.mod_entries = []
        self.mod_data = []
        for x in self.change_stack:
            self.do_change(x, False)

        # check for indicated instance
        if 'subtype' not in d:
            d['subtype'] = 0
        d2 = dict(type=d['type'], group=d['group'], instance=d['instance'], subtype=d['subtype'])
        return d2 in self.mod_entries

    def get(self, i, s):
        """Return data for this instance, including modifications."""
        d = self.last_find.copy()       # save resource data of this instance

        # get current modifications
        self.mod_entries = []
        self.mod_data = []
        for x in self.change_stack:
            self.do_change(x, False)

        # return data/modified data
        d2 = dict(type=d['type'], group=d['group'], instance=d['instance'], subtype=d['subtype'])
        if d2 in self.mod_entries:
            x = self.mod_entries.index(d2)
            return self.mod_data[x]

        return dz_pack.Package.get(self, d['offset'], d['size'])

    def can_write(self):
        return len(self.change_stack) != 0

    def write(self):
        """Save changes to a package."""
        b, f = None, None

        # b = current file data
        try:
            f = open(self.filename, 'rb')
            b = f.read()
        except:
            logging.error("write: file %s failed to read" % self.filename)
            return False

        if f: f.close()

        # build mod lists
        self.mod_entries = []
        self.mod_data = []

        while len(self.change_stack) != 0:
            d = self.change_stack.pop(0)    # start at the bottom and work up
            self.do_change(d)               # apply change

        # apply changes
        b_toc = ""
        b_body = ""
        hdr = self.hdr_size
        ix = 0
        gx = 0
        ex = 0
        t_size = 0

        # rebuild body and toc
        for i in range(self.toc_num):

            # get entry
            xs = self.g_data[gx]
            d2 = dict(type=xs[ex][0], group=self.groups[gx], instance=xs[ex][1], subtype=xs[ex][2])
            o = xs[ex][3]
            s = xs[ex][4]

            # append data to body
            if d2 not in self.mod_entries:
                if o and s:
                    b_body += b[o:o+s]
            else:
                x = self.mod_entries.index(d2)
                b_body += self.mod_data[x]
                s = len(self.mod_data[x])

            # append entry to toc
            if self.toc_ver > 1:
                b_toc += struct.pack('<IIIIII', d2['type'], d2['group'], d2['instance'], d2['subtype'], ix+hdr, s)
                t_size += 24
            else:
                b_toc += struct.pack('<IIIII', d2['type'], d2['group'], d2['instance'], ix+hdr, s)
                t_size += 20

            # next
            ix += s     # current body index
            ex += 1     # next entry
            if ex >= len(xs):
                ex = 0  # first entry in next group
                gx += 1

        # rebuild header
        b_hdr = b[:0x24] + struct.pack('<III', self.toc_num, ix+hdr, t_size)
        b_hdr += b[0x30:hdr]    # note: TOC minor ver is at 0x3C

        # concat all parts
        b2 = b_hdr + b_body + b_toc

        # store
        if b2:
            try:
                f = open(self.filename, 'wb')
                f.write(b2)
            except:
                logging.error("write: write file %s failed" % self.filename)
            if f: f.close()

        dz_pack.Package.reload_toc(self)

        return True

    def build_start(self):
        """Start building a package (note: truncates file)."""
        b, f = None, None

        # b = current file data
        try:
            f = open(self.filename, 'rb')
            b = f.read()
        except:
            logging.error("build_start: file %s failed to read" % self.filename)
            return False

        if f: f.close()

        # start out with header
        b2 = b[:self.hdr_size]

        self.toc_num = 0
        self.toc_valid = False
        self.build_toc = ""
        self.build_ix = 0
        self.build_tsize = 0

        try:
            f = open(self.filename, 'wb')
            f.write(b2)
        except:
            logging.error("build_start: write header to file %s failed" % self.filename)
        if f: f.close()

    def build_append(self, d):
        """Append instance to package build."""
        f = None

        # append data to body
        try:
            f = open(self.filename, 'ab')
            f.write(d['data'])
            s = len(d['data'])
        except:
            logging.error("build_append: write file %s failed" % self.filename)
        if f: f.close()

        # append entry to toc
        if self.toc_ver > 1:
            self.build_toc += struct.pack('<IIIIII', d['type'], d['group'], d['instance'], d['subtype'], \
                self.build_ix+self.hdr_size, s)
            self.build_tsize += 24
        else:
            self.build_toc += struct.pack('<IIIII', d['type'], d['group'], d['instance'], \
                self.build_ix+self.hdr_size, s)
            self.build_tsize += 20

        # next
        self.build_ix += s
        self.toc_num += 1

    def build_end(self):
        """Finish building a package."""
        f = None

        # append toc to body
        try:
            f = open(self.filename, 'ab')
            f.write(self.build_toc)
        except:
            logging.error("build_end: write toc to file %s failed" % self.filename)
        if f: f.close()

        f = None

        # update the header
        try:
            f = open(self.filename, 'r+b')
            b = f.read(self.hdr_size)
            b2 = b[:0x24] + struct.pack('<III', self.toc_num, self.build_ix+self.hdr_size, self.build_tsize)
            b2 += b[0x30:]    # note: TOC minor ver is at 0x3C
            f.seek(0)
            f.write(b2)
        except:
            logging.error("build_end: update header of file %s failed" % self.filename)
        if f: f.close()

        dz_pack.Package.reload_toc(self)

        return True
