# package handling class
# -------------------------------------
# - constructor:
#
# Package(filename, prefs, is_objects): initialize and load a package
#
# - methods:
#
# reload_toc(): loads the package file's table of contents
# get(offset, size): return string data of specified file in array (None = error)
# get2(offset, size_alpha): return truncated data (None = error)
# unpack(data): return unpacked string of data if packed
# unpack2(data): return unpacked string data (checking against buffer overflow)
# find(dict): return the first matching file (None = not found)
# get_name(offset, size): get the title of a found entry (None = missing)
# get_name_lua(offset, size): get the title of a found Lua entry (None = missing)

import cPickle
import logging
import struct

import dz_util
from dz_defs import *

class Package:
    def __init__(self, file_name, prefs, is_objects):
        """Initialize default package variables."""
        self.prefs = prefs              # prefs linkage
        self.is_objects = is_objects    # is this package global?
        self.objects = None             # global linkage (set externally)
        self.group = 0                  # current group
        self.instance = 0               # current instance
        self.subtype = 0                # current subtype (high instance)
        self.semi = 0                   # current semi-global group
        self.pc_size = 50               # 1. change size 2. ??? 3. profit
        self.markup = []                # current markup
        self.m_offset = 0               # current offset for markup
        self.ext_flag = True            # use external refs when dumping
        self.xref_cache = {}
        self.filename = file_name
        self.last_find = None
        self.reload_toc()

    def reload_toc(self):
        """Load the TOC of a given package file."""
        self.pack_cache = {}            # reset the pack cache

        # setup TOC
        self.toc_num = 0                # invalid = default
        self.toc_start = 0
        self.toc_size = 0
        self.toc_ver = 0
        self.toc_valid = False
        self.groups = []
        self.g_data = []

        f = None
        try:
            f = open(self.filename, 'rb')
            f.seek(0x24)
            self.toc_num = dz_util.s2i(f.read(4))
            self.toc_start = dz_util.s2i(f.read(4))
            self.toc_size = dz_util.s2i(f.read(4))
            f.seek(0x3C)
            self.toc_ver = dz_util.s2i(f.read(4))

            if self.toc_size != 0:
                f.seek(self.toc_start)
                self.toc_data = f.read(self.toc_size)
                self.toc_valid = True
        except:
            logging.error("reload_toc: read file %s failed" % self.filename)

        if f:
            f.close()
            if self.toc_valid:
                x = 0
                t_num = 0
                self.groups = []
                self.g_data = []

                if self.is_objects:
                    # try the picked TOC first, if it exists
                    try:
                        f = open("objects-toc", 'r')
                    except:
                        f = None
                        logging.info("objects-toc not found, creating")
                    if f:
                        try:
                            d = cPickle.load(f)     # load prefs dict
                        except:
                            d = None
                            logging.error("objects-toc failed to load")
                        f.close()
                        if d:
                            try:
                                t_num = d['t_num']
                                self.groups = d['groups']
                                self.g_data = d['g_data']
                            except:
                                logging.warning("objects-toc invalidated, reloading from package")

                # this structure is an A+ optimization for find()
                # but takes a bit longer to reload (thus the pickled toc for objects.package)
 
                if t_num != self.toc_num:       # verify objects TOC data with number of entries
                    t = self.toc_data
                    while x < self.toc_size:
                        if self.toc_ver > 1:
                            x2 = struct.unpack('<IIIIII', t[x:x+24])
                            x += 24
                            xt = x2[0]
                            xg = x2[1]
                            xi1 = x2[2]
                            xi2 = x2[3]
                            xo = x2[4]
                            xs = x2[5]
                        else:
                            x2 = struct.unpack('<IIIII', t[x:x+20])
                            x += 20
                            xt = x2[0]
                            xg = x2[1]
                            xi1 = x2[2]
                            xi2 = 0
                            xo = x2[3]
                            xs = x2[4]
                        x3 = [xt,xi1,xi2,xo,xs]
                        if xg not in self.groups:
                            self.groups.append(xg)
                            self.g_data.append([x3])
                        else:
                            x4 = self.groups.index(xg)
                            self.g_data[x4].append(x3)

                    if self.is_objects:
                        d = dict(t_num=self.toc_num, groups=self.groups, g_data=self.g_data)
                        try:
                            f = open("objects-toc", 'w')
                        except:
                            f = None
                            logging.error("objects-toc failed to open for writing")
                        if f:
                            try:
                                cPickle.dump(d, f)       # save toc
                            except:
                                logging.error("objects-toc failed to dump")
                            f.close()

    def get(self, i, s):
        """Load and unpack the data from a package file."""
        b, f = None, None
        if self.toc_valid:

            # check for this in the pack cache, first
            if i in self.pack_cache:
                return self.pack_cache[i]

            try:
                f = open(self.filename, 'rb')
                f.seek(i)
                b = f.read(s)
            except:
                logging.error("get: read file %s failed" % self.filename)
            if f: f.close()

            b = self.unpack(b)

        else:
            logging.error("get: attempted on package with invalid toc")

        # queue this into the pack cache
        if b:
            self.pack_cache.update({i:b[:]})
            if len(self.pack_cache) > self.pc_size:
                self.pack_cache.popitem()   # todo: improve this
        return b

    def get2(self, i, s):
        """Load and unpack data (truncated) from package file."""
        b, f = None, None
        if self.toc_valid:
            try:
                f = open(self.filename, 'rb')
                f.seek(i)
                b = f.read(s)
            except:
                logging.error("get2: read file %s failed" % self.filename)
            if f: f.close()
            return self.unpack2(b)
        else:
            logging.error("get2: attempted on package with invalid toc")
        return None

    def unpack(self, a):
        """Convert full data to its original form, return buffer."""
        b = a

        if b and len(b) >= 10:
            s1 = dz_util.s2i(a[0:4])        # raw size
            if ord(a[4]) == 0x10 and ord(a[5]) == 0xFB:     # check the cookie
                s2 = (ord(a[6]) << 16) + (ord(a[7]) << 8) + ord(a[8])   # unpacked size
                b = ""
                x1, x2, o = 9, 0, 0

                while x1 < s1:              # decode ctrl bytes
                    c = ord(a[x1])
                    x1 += 1
                    if (c & 0x80) == 0:     # 0ooqqqrr oooooooo
                        r = c & 3
                        q = ((c >> 2) & 7) + 3
                        o = ((c << 3) & 0x300) + ord(a[x1])
                        x1 += 1
                    elif (c & 0x40) == 0:   # 10qqqqqq rroooooo oooooooo
                        q = (c & 0x3F) + 4
                        c = ord(a[x1])
                        r = (c >> 6) & 3
                        o = ((c << 8) & 0x3F00) + ord(a[x1+1])
                        x1 += 2
                    elif (c & 0x20) == 0:   # 110oqqrr oooooooo oooooooo qqqqqqqq
                        r = c & 3
                        o = ((c << 12) & 0x10000) + (ord(a[x1]) << 8)
                        o += ord(a[x1+1])
                        q = ((c << 6) & 0x300) + ord(a[x1+2]) + 5
                        x1 += 3
                    else:                   # 111rrrrr / 111111rr
                        r = ((c & 0x1F) + 1) << 2
                        if ((c & 0x1C) == 0x1C):
                            r = c & 3
                        q = 0

                    if x2 + r <= s2:        # range check
                        b += a[x1:x1+r]     # literal copy
                    x1 += r
                    x2 += r

                    o = (x2 - o) - 1        # current index (of last read byte) - offset
                    if x2 + q <= s2:        # range check
                        for i in range(q):
                            b += b[o+i]     # repeat block
                    x2 += q

                if x1 != s1 or x2 != s2:
                    logging.critical("unpack: block size mismatch")

        return b

    def unpack2(self, a):
        """Partially decode a truncated data buffer."""
        b = a

        if b and len(b) >= 10:
            s1 = len(a)                     # actual size
            if ord(a[4]) == 0x10 and ord(a[5]) == 0xFB:     # check the cookie
                s2 = (ord(a[6]) << 16) + (ord(a[7]) << 8) + ord(a[8])   # unpacked size
                b = ""
                x1, x2, o = 9, 0, 0

                while x1 < s1:              # decode ctrl bytes
                    c = ord(a[x1])
                    x1 += 1
                    if (c & 0x80) == 0 and x1 < s1:         # 0ooqqqrr oooooooo
                        r = c & 3
                        q = ((c >> 2) & 7) + 3
                        o = ((c << 3) & 0x300) + ord(a[x1])
                        x1 += 1
                    elif (c & 0x40) == 0 and x1 + 1 < s1:   # 10qqqqqq rroooooo oooooooo
                        q = (c & 0x3F) + 4
                        c = ord(a[x1])
                        r = (c >> 6) & 3
                        o = ((c << 8) & 0x3F00) + ord(a[x1+1])
                        x1 += 2
                    elif (c & 0x20) == 0 and x1 + 2 < s1:   # 110oqqrr oooooooo oooooooo qqqqqqqq
                        r = c & 3
                        o = ((c << 12) & 0x10000) + (ord(a[x1]) << 8)
                        o += ord(a[x1+1])
                        q = ((c << 6) & 0x300) + ord(a[x1+2]) + 5
                        x1 += 3
                    else:                   # 111rrrrr / 111111rr
                        r = ((c & 0x1F) + 1) << 2
                        if ((c & 0x1C) == 0x1C):
                            r = c & 3
                        q = 0

                    if x2 + r <= s2 and x1+r < s1:  # range check
                        b += a[x1:x1+r]             # literal copy
                    x1 += r
                    x2 += r

                    o = (x2 - o) - 1                # current index (of last read byte) - offset
                    if x2 + q <= s2 and x1 < s1:    # range check
                        for i in range(q):
                            b += b[o+i]             # repeat block
                    x2 += q

        return b

    def find(self, d):
        """Search through the TOC of this package."""
        if self.toc_valid:
            # default is "any" and start at index = (0, 0)
            x = [0, 0]
            eSubType = 0
            bg, bt, bi, bs = False, False, False, False

            # determine where to start/what to look for
            if 'index' in d:
                x = d['index']
            if 'group' in d:
                bg = True
                mGroup = d['group']
            if 'type' in d:
                bt = True
                mType = d['type']
            if 'instance' in d:
                bi = True
                mInstance = d['instance']
            if 'subtype' in d:
                bs = True
                mSubType = d['subtype']

            # note: may be a little slow on type-only finds

            i = x[0]
            j = x[1]
            if bg and mGroup not in self.groups:
                return None

            if bg and mGroup in self.groups:
                i2 = self.groups.index(mGroup)
                if i > i2:
                    return None
                elif i != i2:
                    i = i2
                    j = 0

            while i < len(self.groups):
                eGroup = self.groups[i]
                xs = self.g_data[i]         # this group's data sets

                while j < len(xs):
                    eType = xs[j][0]
                    eInstance = xs[j][1]
                    eSubType = xs[j][2]
                    j += 1

                    if bt and mType != eType:
                        continue
                    if bi and mInstance != eInstance:
                        continue
                    if bs and mSubType != eSubType:
                        continue

                    x = [i, j]
                    eOffset = xs[j-1][3]
                    eSize = xs[j-1][4]

                    # return all the results if found, index with next value
                    self.last_find = dict(type=eType, group=eGroup, instance=eInstance, subtype=eSubType, \
                        offset=eOffset, size=eSize, index=x)
                    return self.last_find

                if bg: return None          # group specified? not found
                i += 1                      # next group
                j = 0

        return None

    def get_name(self, o, s):
        """Return the title (without adornments) of a found entry."""
        x = 256         # (this value just needs to be large enough to get the title)
        if x > s: x = s                 # check against actual raw size

        b = self.get2(o, x)             # get data out of a truncated buffer
        if not b: return ""             # inexplicably missing (!)

        j = 0x40                        # get title (cstring, maxchars = 0x40 at index = 0)
        if '\x00' in b[:0x40]: j = b.index('\x00')

        return b[:j]

    def get_name_lua(self, o, s):
        """Return the title (without adornments) of a found Lua entry."""
        x = 256         # (this value just needs to be large enough to get the title)
        if x > s: x = s                 # check against actual raw size

        b = self.get2(o, x)             # get data out of a truncated buffer
        if not b: return ""             # inexplicably missing (!)

        j = dz_util.s2i(b[4:8])         # length
        return b[8:8+j]                 # title follows length
