﻿#!/usr/bin/python
# -*- coding: utf-8 -*-
# misc. utilities

import logging
import os
import struct

import dz_pack
from dz_defs import *

def s2i(s):
    """Convert a data string (little-endian) to an integer."""
    if len(s) == 4:
        x = struct.unpack('<I', s)
        return x[0]
    elif len(s) == 2:
        x = struct.unpack('<H', s)
        return x[0]

    x, m = 0, 0
    for i in range(len(s)):
        x += ord(s[i]) << m
        m += 8
    return x

def calc_crc24(s):
    """Get the 24-bit hash value of a given string."""
    c = CRC24_INIT
    for i in range(len(s)):
        c2 = ord(s[i])
        if c2 >= 0x41 and c2 <= 0x5A: c2 += 0x20  # convert to locase
        c ^= (c2 << 16)
        for j in range(8):
            c <<= 1
            if (c & 0x1000000) != 0: c ^= CRC24_POLY
    return c

def ext16(x):
    """Return a sign-extended 16-bit value string."""
    if x > 0x7FFF: x -= 0x10000
    if x < 16 and x > -16:
        return "%d" % x
    else:
        return "%d (0x%X)" % (x, (x & 0xFFFF))

def hex2int(s):
    """Convert a hexidecimal string to a value."""
    if s[:2] != "0x":
        s = "0x" + s
    return eval(s)

def chk_range(l, x):
    """Return string item from list (or unknown if out of range)."""
    try:
        s = l[x]
    except IndexError:
        s = "unknown 0x%X" % x
        logging.warning("index error in list at 0x%X" % x)
    return s

def do_audit(obj):
    """Look for screwy things like crc collisions."""
    s1 = obj.prefs.object_group
    s2 = obj.prefs.holiday_group
    for i in range(len(s1)):
        x = s1[i]
        if x in s2:
            logging.warning("0x%X in holiday overrides global" % x)

def get_good_string(s):
    try:
        s2 = s.decode("utf_8")      # UTF-8 -> Unicode (raise a decode error if not UTF-8)
        return s
    except:
        logging.warning("string is not UTF-8, converting from cp1252")
        s = s.decode("cp1252")      # Windows 1252 -> Unicode
        return s.encode("utf_8")    # Unicode -> UTF-8

def get_good_filename(s):
    """Return a usable filename."""
    if os.path.isfile(s):
        return s            # file exists, so path must be good
    h, t = os.path.split(s)
    if h == "":
        h = os.getcwd()
    return os.path.join(h, t)

def get_semi(p, g):
    """Return the hash value of this group's semi-global (0 = no semi)."""
    x = 0   # semi = 0 (default)

    if g in GROUP_FIXUP1:                   # fix crc-24 collisions
        j = GROUP_FIXUP1.index(g)
        s = "FUBAR"     # note: FUBAR = name is uncertain
        r = p.find({'group': g, 'type': TYPE_NREF})
        if r:
            i2 = r['instance']
            if i2 in GROUP_FIXUP2[j]:
                g2 = GROUP_FIXUP2[j].index(i2)
                x = GROUP_FIXUP5[j][g2]
    elif g in p.prefs.object_group:
        j = p.prefs.object_group.index(g)
        x = p.prefs.object_semi[j]          # semi = this object's semi-global
    elif g in p.prefs.holiday_group:
        j = p.prefs.holiday_group.index(g)
        x = p.prefs.holiday_semi[j]         # semi = this object's semi-global
    elif g in p.prefs.sg_ids:
        x = g                               # semi = this group

    if g == 0xFFFFFFFF:                     # get custom semi-global
        r = p.find({'type': TYPE_GLOB, 'group': g})
        if r:
            b = p.get(r['offset'], r['size'])
            if b:
                k = ord(b[0x40])            # length (of pstring)
                if ord(b[len(b)-1]) == 0:
                    s1 = b[0x40:len(b)-1]   # cstring, actually
                else:
                    s1 = b[0x41:0x41+k]
                x = calc_crc24(s1) + 0x7F000000
    return x

def get_html(p, t, s, x=None):
    """Return string encoded with HTML markup."""
    if p.prefs.dump_mode == 1:
        if t == TYPE_ANCHOR_GROUP:
            s1 = "<A NAME='GRP_%X'><SPAN CLASS='t_comment'>%s</SPAN></A>" % (x, s)   # x = group
        elif t == TYPE_LINK_GROUP:
            if p.ext_flag and x != p.group:
                s1 = "<A HREF='group-%X.html' TARGET='codeFrame'><SPAN CLASS='t_comment'>%s</SPAN></A>" % (x, s)
            else:
                s1 = "<A HREF='#GRP_%X' ><SPAN CLASS='t_comment'>%s</SPAN></A>" % (x, s)
        elif t == TYPE_ANCHOR_FUNCTION:
            fl = "%X%X" % (x[0], x[1])  # x = (target group, instance)
            s1 = "<A NAME='%s'><SPAN CLASS='t_comment'>%s</SPAN></A>" % (fl, s)
        elif t == TYPE_LINK_FUNCTION:
            fl = "%X%X" % (x[0], x[1])  # x = (target group, instance)
            if p.ext_flag and x[0] != p.group:
                s1 = "<A HREF='group-%X.html#%s' TARGET='codeFrame'><SPAN CLASS='t_function'>%s</SPAN></A>" \
                    % (x[0], fl, s)
            else:
                s1 = "<A HREF='#%s'><SPAN CLASS='t_function'>%s</SPAN></A>" % (fl, s)
        elif t == TYPE_COMMENT:
            s1 = "<SPAN CLASS='t_comment'>%s</SPAN>" % s
        elif t == TYPE_CODE_LINE:
            s1 = "<SPAN CLASS='t_codeLine'>%s</SPAN>" % s
        elif t == TYPE_GOTO_CODE:
            s1 = "<SPAN CLASS='t_gotoCode'>%s</SPAN>" % s
        elif t == TYPE_PROPERTY:
            s1 = "<SPAN CLASS='t_property'>%s</SPAN>" % s
        elif t == TYPE_FUNCTION:
            s1 = "<SPAN CLASS='t_function'>%s</SPAN>" % s
        elif t == TYPE_OPERAND:
            s1 = "<SPAN CLASS='t_operand'>%s</SPAN>" % s
        elif t == TYPE_OPERATOR:
            s1 = "<SPAN CLASS='t_operator'>%s</SPAN>" % s
        elif t == TYPE_DATA:
            s1 = "<SPAN CLASS='t_data'>%s</SPAN>" % s
        return s1
    else:
        return s

def get_lang(n, b):
    # English names of languages
    lang_names_e = [ "English US","English UK","French","German","Italian",\
        "Spanish","Dutch","Danish","Swedish","Norwegian",\
        "Finnish","Hebrew","Russian","Portuguese","Japanese","Polish",\
        "Simplified Chinese","Traditional Chinese","Thai","Korean","Hindi",\
        "Arabic","Bulgarian","Cyrillic","Ukranian","Czech","Greek","Hungarian",\
        "Icelandic","Romanian","Latin","Slovak","Albanian","Turkish",\
        "Portuguese (Brazil)","Swiss French","Canadian French","Belgian French",\
        "Swiss German","Swiss Italian","Belgian Dutch","Mexican Spanish",\
        "Tagalog","Vietnamese" ]

    # actual names of languages
    lang_names_x = [ "English (US)","English (UK)","Français","Deutsch","Italiano",\
        "Español","Nederlands","Dansk","Svenska","Norsk",\
        "Suomi","עברית","Русский","Português","日本語","Polski",\
        "简体字","簡體字","ภาษาไทย","한국어","हिन्दी","العربية","Български","Cyrillic (?)",\
        "Українська","Čeština","Ελληνικά","Magyar","Íslenska","Română",\
        "Latina","Slovenčina","Shqip","Türkçe","Português (Brasil)","Français (Suisse)",\
        "Français (Canada)","Français (Belgique)","Deutsch (Schweiz)",\
        "Italiano (Svizzera)","Nederlands (België)","Español (México)",\
        "Tagalog","Tiếng Việt" ]

    if n > len(lang_names_e):
        return "unknown %d" % n

    if b:
        return lang_names_x[n-1]
    else:
        return lang_names_e[n-1]
